import re

class Tag:
    no_quoting = re.compile(r'^[a-z0-9\.\*_-]+$', re.I)

    def __init__(self, tag, *params):
        self.tag = tag
        self.params = list(params)

    def append(self, *contents):
        self.params += list(contents)

    def __iter__(self):
        yield "("+self.tag
        for p in self.params:
            if isinstance(p, Tag):
                for pp in p:
                    yield pp
            else:
                pp = str(p)
                if no_quoting.match():
                    yield pp
                else:
                    yield repr(p)
        yield ")"

    def code(self):
        indent = 0
        ret = ""
        line = ""
        for statement in self:
            assert statement
            line += statement
            if len(line) >= 72:
                ret += "\n" + " " * indent + line
                line = ""
            if statement.startswith("("):
                indent += 2
            if statement.endswith(")"):
                indent -= 2
        if line:
            ret += "\n" + " " * indent + line
            line = ""
        return ret

class Module:
    def __init__(self, name, descr, mode = "smd"):
        self.name = name
        self.descr = descr
        self.mode = mode
        self.contents = []

    def append(self, *objs):
        self.contents += list(objs)

    def code(self):
        root = Tag("module", self.name,
                   Tag("layer", "F.Cu"),
                   Tag("descr", self.descr),
                   Tag("attr", self.mode),
                   )

        for c in self.contents:
            root.append(c.tag())

        return root.code()

class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

class Bbox:
    def __init__(self, top, right, bottom, left):
        self.top = top
        self.right = right
        self.bottom = bottom
        self.left = left

    def __add__(self, other):
        return Bbox(min(self.bottom, other.bottom, self.top, other.top),
                    max(self.right, other.right, self.left, other.left),
                    max(self.bottom, other.bottom, self.top, other.top),
                    min(self.right, other.right, self.left, other.left))

    @property
    def tl(self):
        return Point(self.left, self.top)

    @property
    def tr(self):
        return Point(self.right, self.top)

    @property
    def bl(self):
        return Point(self.left, self.bottom)

    @property
    def br(self):
        return Point(self.right, self.bottom)

class Grob:
    def __init__(self, layer):
        self.layer = layer

    def bbox(self):
        return Bbox(0, 0, 0, 0)

class Text(Grob):
    def __init__(self, layer, position, text, type,
                 line_height = 1.5, char_width = 1.5, stroke_width = 0.3):
        Grob.__init__(self, layer)
        self.position = position
        self.text = text
        self.type = type
        self.line_height = line_height
        self.char_width = char_width
        self.stroke_width = stroke_width

    def kicad_mod(self):
        return Tag("fp_text", self.type, text,
                   Tag("at", self.position.x, self.position.y),
                   Tag("layer", self.layer),
                   Tag("effects", Tag("font", Tag("size",
