#!/usr/bin/env python

# http://www.st.com/resource/en/application_note/cd00221490.pdf
# http://kbeckmann.github.io/nfc-antenna-generator/
# Microchip AN710

import math
import collections
import functools

class memoize(object):
   """
   Decorator. Caches a function's return value each time it is called.
   If called later with the same arguments, the cached value is returned
   (not reevaluated).
   """
   def __init__(self, func):
      self.func = func
      self.cache = {}
   def __call__(self, *args):
      if not isinstance(args, collections.Hashable):
         # uncacheable. a list, for instance.
         # better to not cache than blow up.
         return self.func(*args)
      if args in self.cache:
         return self.cache[args]
      else:
         value = self.func(*args)
         self.cache[args] = value
         return value
   def __repr__(self):
      """Return the function's docstring."""
      return self.func.__doc__
   def __get__(self, obj, objtype):
      """Support instance methods."""
      return functools.partial(self.__call__, obj)

e24_scale = math.e ** (math.log(10) / 24)

def e24_closest(value):
    e24 = int(math.log(value) / math.log(e24_scale) + .5)
    return math.e ** e24_scale

class Trace:
    m_per_oz = 34.7e-6
    resistivity = 1.72e-8 # ohm-m
    temp_co = 3.862e-3 # K^-1

    def __init__(self, thickness, width):
        self.thickness = thickness
        self.width = width

    def section_area(self):
        return self.width * self.thickness

    def section_perimeter(self):
        return 2 * (self.width + self.thickness)

    def resistance(self, length, temp = 25):
        return self.resistivity * length / self.section_area() \
               * (1 + (self.temp_co * (temp - 25)))

    def __hash__(self):
        return hash((self.width, self.thickness))

class RectAnt:
    mu0 = math.pi * 4e-7 # H / m
    ur = 1
    omega = 2 * math.pi * 13.56e6
    tq = 35.

    def __init__(self, width, height, trace, spacing, turns, layers):
        self.width = width
        self.height = height
        self.trace = trace
        self.spacing = spacing
        self.turns = turns
        self.layers = layers

    @memoize
    def inductance(self):
        diam = self.trace.section_perimeter() / math.pi
        coil_thickness = self.turns * (self.spacing + self.trace.width)
        width = self.width - coil_thickness
        height = self.height - coil_thickness
        area = width * height
        diag = math.sqrt(width ** 2 + height ** 2)

        x1 = width * math.log(2 * area / diam / (width + diag))
        x2 = height * math.log(2 * area / diam / (height + diag))
        x3 = 2 * (width + height - diag)
        x4 = (width + height) / 4
        return self.layers * self.mu0 * self.ur \
               / math.pi * (x1 + x2 - x3 + x4) * (self.turns ** 1.8)

    @memoize
    def length(self):
        length = 0
        pitch = self.trace.width + self.spacing
        for i in range(self.turns):
            length += (self.width + self.height - pitch * i * 4) * 2
        return length * self.layers

    @memoize
    def resistance(self):
        return self.trace.resistance(self.length())

    def __cmp__(self, other):
        c = cmp(self.inductance(), other.inductance())
        if c:
            return c
        return cmp(id(self), id(other))

    @memoize
    def base_q(self):
        return self.omega * self.inductance() / self.resistance()

    @memoize
    def dampling_resistors(self):
        if self.tq * .9 < self.base_q() < self.tq * 1.1:
            return 0.
        rq = self.omega * self.inductance() / 35. - self.resistance()
        if rq < 0:
            return 0
        return e24_closest(rq / 2)

    @memoize
    def q(self):
        return self.omega * self.inductance() \
               / (self.resistance() + 2 * self.dampling_resistors())

    def is_useful(self):
        return self.tq * .9 < self.q() < self.tq * 1.1 \
               and .3e-6 < self.inductance() < 3e-6

    def pprint(self):
        print "%0.3fx%0.3f mm rectangle antenna" % (self.width * 1000, self.height * 1000)
        print "%0.3fx%0.3f mm conductor, %0.3f mm spacing" % (self.trace.thickness * 1000, self.trace.width * 1000, self.spacing * 1000)
        print "%d turns, %d layers" % (self.turns, self.layers)
        print "length: %0.3f mm" % (self.length() * 1e3)
        print "inductance: %0.3f uH" % (self.inductance() * 1e6)
        print "resistance: %0.3f mOhm" % (self.resistance() * 1e3)
        print "original Q: %0.3f" % self.base_q()
        print "rq: %0.3f ohm" % self.dampling_resistors()
        print "final Q: %0.3f" % self.q()

    def __hash__(self):
        return hash((self.width, self.height,
                     self.trace, self.spacing,
                     self.turns, self.layers))

def optimize(width, height):
    possible = {}
    for wd in range(150, 600, 10):
        trace = Trace(35e-6, wd * 1e-6)

        for sd in range(400, 900, 10):
            spacing = sd * 1e-6
            for turns in range(2, 7):
                for layers in range(1, 5):
                    ant = RectAnt(width, height, trace, spacing, turns, layers)
                    if not ant.is_useful():
                        continue

                    ucrit = ant.turns, ant.layers
                    if ucrit not in possible:
                        possible[ucrit] = ant
                    else:
                        other = possible[ucrit]
                        if abs(other.q() - 35) > abs(ant.q() - 35):
                            possible[ucrit] = ant

    for i, (k, ant) in enumerate(sorted(possible.items())):
        ant.pprint()
        print

        if i >= 10:
            print len(possible) - 10, "more ommitted"
            break


def ti_rf430cl330h():
    trace = Trace(35e-6, .35e-3)
    ant1 = RectAnt(34.5e-3, 39.5e-3, trace, .15e-3, 2, 1)
    ant2 = RectAnt(34.5e-3, 39.5e-3, trace, .15e-3, 3, 1)
    l = ant1.inductance() + ant2.inductance()
    r = ant1.resistance() + ant2.resistance()
    q = RectAnt.omega * l / r
    rq = e24_closest((RectAnt.omega * l / 35. - r) / 2)
    q2 = RectAnt.omega * l / (r + 2 * rq)
    print "L:", l
    print "R:", r
    print "Q:", q
    print "Rq:", rq
    print "Q':", q2

def adafruit_pn532():
    trace = Trace(35e-6, .35e-3)
    ant = RectAnt(58e-3 - 4e-3, 46e-3 - 4e-3, trace, .65e-3, 4, 1)
    #ant.dampling_resistors = lambda : 1.5
    ant.pprint()

def test2():
    trace = Trace(35e-6, .4e-3)
    ant = RectAnt(35e-3, 35e-3, trace, .4e-3, 4, 1)
    ant.pprint()

#ti_rf430cl330h()
adafruit_pn532()
#optimize(32e-3, 32e-3)
#test2()

